@@ -17,18 +17,18 @@ a full implementation of the training routine given in the paper, with both
 speed-up tricks.
 
 """
-import functools
 import json
 import math
 import multiprocessing as mp
 import os
 import pickle
 import shutil
-from contextlib import ExitStack
-from multiprocessing import Process, sharedctypes
-from operator import mul
+from multiprocessing import Process
+import random
 
 import numpy as np
+from gaussianlda.mp_utils import GaussianLock, SharedArray, TwoSidedLock
+from numpy.core.umath import isinf
 from numpy.linalg import slogdet
 from scipy.linalg import solve_triangular
 from scipy.special import gammaln
@@ -38,11 +38,33 @@ from gaussianlda.utils import get_logger, get_progress_bar, chol_rank1_downdate,
 
 
 class GaussianLDAAliasTrainer:
-    def __init__(self, corpus, vocab_embeddings, vocab, num_tables, alpha, kappa=0.1, log=None, save_path=None,
-                 show_topics=None, mh_steps=2):
+    def __init__(self, corpus, vocab_embeddings, vocab, num_tables, alpha=None, kappa=0.1, log=None, save_path=None,
+                 show_topics=None, mh_steps=2, num_words_for_formatting=None, das_normalization=True, show_progress=True):
+        """
+
+        :param corpus:
+        :param vocab_embeddings:
+        :param vocab:
+        :param num_tables:
+        :param alpha: Dirichlet concentration. Defaults to 1/num_tables
+        :param kappa:
+        :param log:
+        :param save_path:
+        :param show_topics:
+        :param mh_steps:
+        :param num_words_for_formatting: By default, each topic is formatted by computing the probability of
+            every word in the vocabulary under that topic. This can take a long time for a large vocabulary.
+            If given, this limits the number considered to the first
+            N in the vocabulary (which makes sense if the vocabulary is ordered with most common words first).
+        :param das_normalization: Use the normalization of probability distributions used by Das, Zaheer and Dyer's
+            original implementation when computing the sampling probability to choose whether to use the document
+            posterior or language model part of the topic posterior. If False, do not normalize in this way, but use
+            an alternative, which looks to me like it's more correct mathematically.
+        """
         if log is None:
             log = get_logger("GLDA")
         self.log = log
+        self.show_progress = show_progress
         # Vocab is used for outputting topics
         self.vocab = vocab
         self.show_topics = show_topics
@@ -52,11 +74,16 @@ class GaussianLDAAliasTrainer:
         self.mh_steps = mh_steps
 
         # Dirichlet hyperparam
+        if alpha is None:
+            alpha = 1. / num_tables
         self.alpha = alpha
 
+        self.das_normalization = das_normalization
+
         # dataVectors
         self.vocab_embeddings = vocab_embeddings
         self.embedding_size = vocab_embeddings.shape[1]
+        self.num_terms = vocab_embeddings.shape[0]
         # List of list of ints
         self.corpus = corpus
         # numIterations
@@ -98,19 +125,19 @@ class GaussianLDAAliasTrainer:
         # Used in calculate_table_params()
         self.k0mu0mu0T = self.prior.kappa * np.outer(self.prior.mu, self.prior.mu)
 
+        self.num_words_for_formatting = num_words_for_formatting
+
+        self.aliases = VoseAliases.create(self.num_terms, self.num_tables)
+
         self.log.info("Initializing assignments")
         self.initialize()
 
-        self.aliases = [
-            VoseAlias.create(self.num_tables) for i in range(self.vocab_embeddings.shape[0])
-        ]
-
     def initialize(self):
         """
         Initialize the gibbs sampler state.
-        
+
         I start with log N tables and randomly initialize customers to those tables.
-        
+
         """
         # First check the prior degrees of freedom.
         # It has to be >= num_dimension
@@ -141,7 +168,7 @@ class GaussianLDAAliasTrainer:
 
         # Randomly assign customers to tables
         self.table_assignments = []
-        pbar = get_progress_bar(len(self.corpus), title="Initializing")
+        pbar = get_progress_bar(len(self.corpus), title="Initializing", show_progress=self.show_progress)
         for doc_num, doc in enumerate(pbar(self.corpus)):
             tables = list(np.random.randint(self.num_tables, size=len(doc)))
             self.table_assignments.append(tables)
@@ -290,17 +317,31 @@ class GaussianLDAAliasTrainer:
                     don't have to update the parameters
                 else update params of the old table.
         """
+        if self.show_topics is not None:
+            print("Topics after initialization")
+            print(self.format_topics())
+            # Compute the overall usage of topics across the training corpus
+            topic_props = self.table_counts_per_doc.sum(axis=1).astype(np.float64)
+            topic_props /= topic_props.sum()
+            print("Words using topics: {}".format(
+                ", ".join("{}={:.1f}%".format(i, prop) for i, prop in enumerate(topic_props*100.))))
+            topic_doc_props = (self.table_counts_per_doc > 0).astype(np.float64).sum(axis=1)
+            topic_doc_props /= self.num_documents
+            print("Docs using topics: {}".format(
+                ", ".join("{}={:.1f}%".format(i, prop) for i, prop in enumerate(topic_doc_props*100.))))
+
         with VoseAliasUpdater(
                 self.aliases, self.vocab_embeddings,
                 self.prior.kappa, self.prior.nu,
                 self.table_counts, self.table_means, self.table_cholesky_ltriangular_mat,
-                self.log_determinants
+                self.log_determinants, das_normalization=self.das_normalization,
         ) as alias_updater:
             for iteration in range(num_iterations):
+                stats = SamplingDiagnostics()
                 self.log.info("Iteration {}".format(iteration))
 
                 alias_updater.unpause()
-                pbar = get_progress_bar(len(self.corpus), title="Sampling")
+                pbar = get_progress_bar(len(self.corpus), title="Sampling", show_progress=self.show_progress)
                 for d, doc in enumerate(pbar(self.corpus)):
                     if self.show_topics is not None and self.show_topics > 0 and d % self.show_topics == 0:
                         print("Topics after {:,} docs".format(d))
@@ -332,66 +373,138 @@ class GaussianLDAAliasTrainer:
                         # Under the alias method, we only do the full likelihood computation for topics
                         # that already have a non-zero count in the current document
                         non_zero_tables = np.where(self.table_counts_per_doc[:, d] > 0)[0]
-                        # We only compute the posterior for these topics
-                        posterior = np.zeros(len(non_zero_tables), dtype=np.float32)
-                        for nz_table, table in enumerate(non_zero_tables):
-                            log_likelihood = self.log_multivariate_tdensity(x, table)
-                            posterior[nz_table] = np.log(self.table_counts_per_doc[table, d]) + log_likelihood
+                        if len(non_zero_tables) == 0:
+                            # If there's only one word in a doc, there are no topics to compute the full posterior for
+                            no_non_zero = True
+                        else:
+                            no_non_zero = False
+                            # We only compute the posterior for these topics
+                            log_priors = np.log(self.table_counts_per_doc[non_zero_tables, d])
+                            log_likelihoods = np.zeros(len(non_zero_tables), dtype=np.float32)
+                            for nz_table, table in enumerate(non_zero_tables):
+                                log_likelihoods[nz_table] = self.log_multivariate_tdensity(x, table)
+                            log_posterior = log_priors + log_likelihoods
 
-                        # To prevent overflow, subtract by log(p_max)
-                        posterior -= posterior.max()
-                        posterior = np.exp(posterior)
-                        psum = posterior.sum()
-                        posterior /= psum
+                            # To prevent overflow, subtract by log(p_max)
+                            max_log_posterior = log_posterior.max()
+                            scaled_posterior = log_posterior - max_log_posterior
+                            if self.das_normalization:
+                                # Not doing this now, but following what the Java impl does, however odd that seems
+                                psum = np.sum(np.exp(scaled_posterior))
+                            else:
+                                # Java impl subtracts max before computing psum, but this seems to be wrong
+                                # We still subtract first, but then multiply by the max prob afterwards
+                                psum = np.exp(np.log(np.sum(np.exp(scaled_posterior))) + max_log_posterior)
+                            # Now just use the scaled log posterior in the same way as in the Java impl
+                            # They have a bin-search method for sampling from the cumulative dist,
+                            # but we simply normalize and use Numpy to sample
+                            unnormed_posterior = np.exp(scaled_posterior)
+                            normed_posterior = unnormed_posterior / unnormed_posterior.sum()
 
                         # Don't let the alias parameters get updated in the middle of the sampling
-                        alias = self.aliases[cust_id]
-                        with alias.lock:
-                            select_pr = psum / (psum + self.alpha*alias.wsum)
+                        self.aliases.lock.acquire_read(cust_id)
+                        select_pr = psum / (psum + self.alpha*self.aliases.likelihood_sum.np[cust_id])
 
-                            # MHV to draw new topic
-                            # Take a number of Metropolis-Hastings samples
-                            for r in range(self.mh_steps):
-                                # 1. Flip a coin
-                                if np.random.sample() < select_pr:
-                                    # Choose from the computed posterior dist
-                                    temp = np.random.choice(len(non_zero_tables), p=posterior)
-                                    new_table_id = non_zero_tables[temp]
-                                else:
-                                    new_table_id = alias.sample_vose()
+                        # MHV to draw new topic
+                        # Take a number of Metropolis-Hastings samples
+                        current_sample = old_table_id
+                        # Calculate the true likelihood of this word under the current sample,
+                        # for calculating acceptance prob
+                        current_sample_log_prob = self.log_multivariate_tdensity(x, current_sample)
+                        for r in range(self.mh_steps):
+                            # 1. Flip a coin
+                            if not no_non_zero and np.random.sample() < select_pr:
+                                # Choose from the exactly computed posterior dist, only allowing
+                                # topics already sampled in the doc
+                                temp = np.random.choice(len(non_zero_tables), p=normed_posterior)
+                                new_sample = non_zero_tables[temp]
+                                stats.log_select_pr(True, select_pr)
+                            else:
+                                # Choose from the alias, allowing any topic but using slightly
+                                # out-of-date likelihoods
+                                new_sample = self.aliases.sample_vose(cust_id)
+                                stats.log_select_pr(False, select_pr)
 
-                                if new_table_id != old_table_id:
-                                    # 2. Find acceptance probability
-                                    old_prob = self.log_multivariate_tdensity(x, old_table_id)
-                                    new_prob = self.log_multivariate_tdensity(x, new_table_id)
+                            if new_sample != current_sample:
+                                # 2. Find acceptance probability
+                                new_sample_log_prob = self.log_multivariate_tdensity(x, new_sample)
+                                # This can sometimes generate an overflow warning from Numpy
+                                # We don't care, though: in that case acceptance > 1., so we always accept
+                                with np.errstate(over="ignore"):
+                                    # From my reading of:
+                                    # Li et al. (2014): Reducing the sampling complexity of topic models
+                                    # the acceptance probability should be as follows:
                                     acceptance = \
-                                        (self.table_counts_per_doc[new_table_id, d] + self.alpha) / \
-                                        (self.table_counts_per_doc[old_table_id, d] + self.alpha) * \
-                                        np.exp(new_prob - old_prob) * \
-                                        (self.table_counts_per_doc[old_table_id, d]*old_prob +
-                                         self.alpha*alias.w.np[old_table_id]) / \
-                                        (self.table_counts_per_doc[new_table_id, d]*new_prob +
-                                         self.alpha*alias.w.np[new_table_id])
-                                    # 3. Compare against uniform[0,1]
-                                    if np.random.sample() < acceptance:
-                                        # This seems puzzling, but follows the Java impl exactly...
-                                        old_table_id = new_table_id
+                                        (self.table_counts_per_doc[new_sample, d] + self.alpha) / \
+                                        (self.table_counts_per_doc[current_sample, d] + self.alpha) * \
+                                        np.exp(new_sample_log_prob - current_sample_log_prob) * \
+                                        (self.table_counts_per_doc[current_sample, d]*np.exp(current_sample_log_prob) +
+                                         self.alpha*np.exp(self.aliases.log_likelihoods.np[cust_id, current_sample])) / \
+                                        (self.table_counts_per_doc[new_sample, d]*np.exp(new_sample_log_prob) +
+                                         self.alpha*np.exp(self.aliases.log_likelihoods.np[cust_id, new_sample]))
+                                    # The Java implementation, however, does this:
+                                    #acceptance = \
+                                    #    (self.table_counts_per_doc[new_table_id, d] + self.alpha) / \
+                                    #    (self.table_counts_per_doc[current_sample, d] + self.alpha) * \
+                                    #    np.exp(new_prob - old_prob) * \
+                                    #    (self.table_counts_per_doc[current_sample, d]*old_log_prob +
+                                    #     self.alpha*alias.w.np[current_sample]) / \
+                                    #    (self.table_counts_per_doc[new_table_id, d]*new_log_prob +
+                                    #     self.alpha*alias.w.np[new_table_id])
+                                    # The difference is the Java impl doesn't exp the log likelihood in the last
+                                    # fraction, i.e. it uses a log prob instead of a prob
+                                # 3. Compare against uniform[0,1]
+                                # If the acceptance prob > 1, we always accept: this means the new sample
+                                # has a higher probability than the old
+                                if isinf(acceptance) or acceptance >= 1. or np.random.sample() < acceptance:
+                                    # No need to sample if acceptance >= 1
+                                    # If the acceptance prob < 1, sample whether to accept or not, such that
+                                    # the more likely the new sample is compared to the old, the more likely we
+                                    # are to keep it
+                                    current_sample = new_sample
+                                    current_sample_log_prob = new_sample_log_prob
+                                    stats.log_acceptance(True, acceptance)
+                                else:
+                                    stats.log_acceptance(False, acceptance)
+                                # NOTE: There seems to be a small error in the Java implementation here
+                                # On the last MH step, it doesn't make any difference whether we accept the
+                                # sample or not - we always end up using it
+                        self.aliases.lock.release_read()
 
                         # Now have a new assignment: add its counts
-                        self.table_assignments[d][w] = new_table_id
+                        self.table_assignments[d][w] = current_sample
                         with self.table_counts.lock:
-                            self.table_counts.np[new_table_id] += 1
-                        self.table_counts_per_doc[new_table_id, d] += 1
-                        self.sum_squared_table_customers[new_table_id] += np.outer(x, x)
+                            self.table_counts.np[current_sample] += 1
+                        self.table_counts_per_doc[current_sample, d] += 1
+                        self.sum_squared_table_customers[current_sample] += np.outer(x, x)
 
-                        self.update_table_params(new_table_id, cust_id)
+                        self.update_table_params(current_sample, cust_id)
 
                 # Pause the alias updater until we start the next iteration
                 alias_updater.pause()
 
+                # Output some useful stats about sampling
+                if stats.acceptance_used():
+                    self.log.info("Acceptance rate = {:.2f}%, mean acceptance: {:.2f}".format(
+                        stats.acceptance_rate()*100., stats.mean_acceptance()))
+                else:
+                    self.log.info("No new samples drawn")
+                self.log.info("Prior select rate = {:.2f}%, mean select_pr = {:.2f}".format(
+                    stats.select_pr_rate() * 100., stats.mean_select_pr()
+                ))
+
                 if self.show_topics is not None:
                     print("Topics after iteration {}".format(iteration))
                     print(self.format_topics())
+                    # Compute the overall usage of topics across the training corpus
+                    topic_props = self.table_counts_per_doc.sum(axis=1).astype(np.float64)
+                    topic_props /= topic_props.sum()
+                    print("Words using topics: {}".format(
+                        ", ".join("{}={:.1f}%".format(i, prop) for i, prop in enumerate(topic_props*100.))))
+                    topic_doc_props = (self.table_counts_per_doc > 0).astype(np.float64).sum(axis=1)
+                    topic_doc_props /= self.num_documents
+                    print("Docs using topics: {}".format(
+                        ", ".join("{}={:.1f}%".format(i, prop) for i, prop in enumerate(topic_doc_props*100.))))
 
                 if self.save_path is not None:
                     self.log.info("Saving model")
@@ -401,10 +514,16 @@ class GaussianLDAAliasTrainer:
         if topics is None:
             topics = list(range(self.num_tables))
 
+        if self.num_words_for_formatting is not None:
+            # Limit to the first N words to consider for inclusion in a topic's representation
+            embeddings = self.vocab_embeddings[:self.num_words_for_formatting]
+        else:
+            embeddings = self.vocab_embeddings
+
         topic_fmt = []
         for topic in topics:
             # Compute the density for all words in the vocab
-            word_scores = self.log_multivariate_tdensity(self.vocab_embeddings, topic)
+            word_scores = self.log_multivariate_tdensity(embeddings, topic)
             word_probs = np.exp(word_scores - word_scores.max())
             word_probs /= word_probs.sum()
             topic_fmt.append(
@@ -443,57 +562,78 @@ class GaussianLDAAliasTrainer:
                 pickle.dump(data, f)
 
 
-class VoseAlias:
+class VoseAliases:
     """
-    Implementation of Vose alias for faster sampling.
+    Implementation of Vose alias for faster sampling. This implementation
+    stores all the aliases for the entire vocabulary in a single data
+    structure, reducing memory overheads and making synchronization more
+    efficient.
+
+    The old implementation, where a separate table was maintained for every
+    word, each with a separate lock, is in `._old_alias.py` for reference.
 
     See https://github.com/rajarshd/Gaussian_LDA/blob/master/src/util/VoseAlias.java.
 
-    Each alias has its own lock. You should make sure you acquire the lock while either
-    using the alias for sampling or updating lock parameters, to ensure that parameter
-    updates and sampling using the parameters don't get interleaved.
+    We enforce locking with a single lock that ensures that reading and
+    writing of the same word's alias don't overlap. It's assumed that there's
+    exactly one process reading and one writing, as in the Gaussian LDA
+    sampler.
 
     """
-    def __init__(self, alias, prob, w, lock):
-        self.n = alias.shape
+    def __init__(self, alias, prob, log_likelihoods, likelihood_sum, lock):
+        self.n = alias.shape[1]
         # Alias indices
         self.alias = alias
         # Contains proportions and alias probabilities
         self.prob = prob
-        self.w = w
+        # Contains log likelihoods of the word given each topic
+        self.log_likelihoods = log_likelihoods
+        self.likelihood_sum = likelihood_sum
+
+        self.gen = random.Random()
 
-        # Make sure we don't mix sampling and updating
         self.lock = lock
 
     @staticmethod
-    def create(num):
-        alias = SharedArray.create(num, "int")
-        prob = SharedArray.create(num, "float")
-        w = SharedArray.create(num, "float")
-        lock = mp.Lock()
-        return VoseAlias(alias, prob, w, lock)
+    def create(num_words, num_topics):
+        # Create big arrays to hold all the values for all words
+        alias = SharedArray.create((num_words, num_topics), "int")
+        prob = SharedArray.create((num_words, num_topics), "float")
+        log_likelihoods = SharedArray.create((num_words, num_topics), "float")
+        likelihood_sum = SharedArray.create(num_words, "float")
+        lock = TwoSidedLock.create()
+        return VoseAliases(alias, prob, log_likelihoods, likelihood_sum, lock)
 
     def __getstate__(self):
-        return self.alias, self.prob, self.w, self.lock
+        return self.alias, self.prob, self.log_likelihoods, self.likelihood_sum, self.lock
 
     def __setstate__(self, state):
         self.__init__(*state)
 
-    @property
-    def wsum(self):
-        return self.prob.np.sum()
-
-    def sample_vose(self):
-        # 1. Generate a fair die roll from an n-sided die; call the side i
-        fair_die = np.random.randint(self.n)
+    def sample_vose(self, word):
+        # 1. Generate a fair die roll from an n-sided die
+        fair_die = self.gen.randint(0, self.n-1)
         # 2. Flip a biased coin that comes up heads with probability Prob[i]
-        m = np.random.sample()
+        m = self.gen.random()
         # 3. If the coin comes up "heads," return i. Otherwise, return Alias[i]
-        if m > self.prob.np[fair_die]:
-            return self.alias.np[fair_die]
+        if m > self.prob.np[word, fair_die]:
+            return self.alias.np[word, fair_die]
         else:
             return fair_die
 
+    def sample_numpy(self, word):
+        """
+        Draw sample without using the Vose alias, instead just using Numpy's sample method.
+        This should be less efficient, presumably, but is guaranteed to sample correctly,
+        so I'm using it as debugging.
+
+        """
+        lp = self.log_likelihoods.np[word, :]
+        lp -= lp.max()
+        p = np.exp(lp)
+        p /= p.sum()
+        return np.random.choice(len(p), p=p)
+
 
 class VoseAliasUpdater(Process):
     """
@@ -502,9 +642,10 @@ class VoseAliasUpdater(Process):
 
     """
     def __init__(self, aliases, embeddings, kappa, nu, table_counts, table_means, table_cholesky_ltriangular_mat,
-                 log_determinants):
+                 log_determinants, das_normalization=True):
         super().__init__()
         self.aliases = aliases
+        self.das_normalization = das_normalization
 
         # Gaussian parameters
         self.log_determinants = log_determinants.np
@@ -562,17 +703,26 @@ class VoseAliasUpdater(Process):
                 # This is because when we will be normalizing after exponentiating, each entry will be
                 # exp(log p_i - log p_max )/\Sigma_i exp(log p_i - log p_max)
                 # The log p_max cancels put and prevents overflow in the exponentiating phase
-                log_likelihoods -= log_likelihoods.max()
+                ll_max = log_likelihoods.max()
+                log_likelihoods_scaled = log_likelihoods - ll_max
                 # Exp them all to get the new w array
-                w = np.exp(log_likelihoods)
+                w = np.exp(log_likelihoods_scaled)
+                if self.das_normalization:
+                    # Not doing this now, but following what the Java impl does, however odd that seems
+                    likelihood_sum = w.sum()
+                else:
+                    # Sum of likelihood has to be scaled back to its original sum, before we subtracted the max log
+                    likelihood_sum = np.exp(np.log(w.sum()) + ll_max)
                 # Update the alias and probs using this w
                 new_alias, new_prob = self.generate_table(w)
                 # Update the parameters of this word's alias
-                term_alias = self.aliases[term_id]
-                with term_alias.lock:
-                    np.copyto(term_alias.w.np, w)
-                    np.copyto(term_alias.alias.np, new_alias)
-                    np.copyto(term_alias.prob.np, new_prob)
+                # Acquire the lock to write to this word
+                self.aliases.lock.acquire_write(term_id)
+                np.copyto(self.aliases.log_likelihoods.np[term_id], log_likelihoods)
+                np.copyto(self.aliases.alias.np[term_id], new_alias)
+                np.copyto(self.aliases.prob.np[term_id], new_prob)
+                self.aliases.likelihood_sum.np[term_id] = likelihood_sum  # This one is a single value
+                self.aliases.lock.release_write()
 
             # If this was the first iteration, we tell the main process that we've finished it
             self.initialized.set()
@@ -592,6 +742,7 @@ class VoseAliasUpdater(Process):
         self.running.set()
 
     def generate_table(self, w):
+        # This implementation is copied from Java GaussianLDA
         # Generate a new prob array and alias table
         alias = np.zeros(self.num_tables, dtype=np.int32)
         # Contains proportions and alias probabilities
@@ -602,9 +753,15 @@ class VoseAliasUpdater(Process):
         p = (w * self.num_tables) / w.sum()
         # 3. For each scaled probability pi:
         #   a. If pi<1, add i to Small.
-        #   b. Otherwise(pi≥1), add i to Large.
-        small = list(np.where(p < 1.)[0])
-        large = list(np.where(p > 1.)[0])
+        #   b. Otherwise(pi>=1), add i to Large.
+        small, large = [], []
+        for i, pi in enumerate(p):
+            if pi < 1.:
+                small.append(i)
+            else:
+                large.append(i)
+        #small = list(np.where(p < 1.)[0])
+        #large = list(np.where(p >= 1.)[0])
         # 4. While Small and Large are not empty : (Large might be emptied first)
         #    a. Remove the first element from Small; call it l.
         #    b. Remove the first element from Large; call it g.
@@ -638,7 +795,7 @@ class VoseAliasUpdater(Process):
 
         return alias, prob
 
-    def log_multivariate_tdensity_tables(self, x):
+    def __log_multivariate_tdensity_tables(self, x):
         """
         A local version of the likelihood function from the main model, using the
         copy of the parameters we have in our local process.
@@ -677,81 +834,93 @@ class VoseAliasUpdater(Process):
                   )
         return logprob
 
-
-class GaussianLock(ExitStack):
-    """
-    Lock all the parameters of a gaussian, so we can either update them without them getting
-    used in a partially updated state or use them without risking some getting updated.
-
-    The arguments are SharedArrays. All of their locks will be acquired when entering
-    the context manager.
-
-    """
-    def __init__(self, table_counts, table_means, table_cholesky_ltriangular_mat, log_determinants):
-        super().__init__()
-        self.log_determinants_lock = log_determinants.lock
-        self.table_cholesky_ltriangular_mat_lock = table_cholesky_ltriangular_mat.lock
-        self.table_means_lock = table_means.lock
-        self.table_counts_lock = table_counts.lock
-
-    def __enter__(self):
-        obj = super().__enter__()
-        obj.enter_context(obj.log_determinants_lock)
-        obj.enter_context(obj.table_cholesky_ltriangular_mat_lock)
-        obj.enter_context(obj.table_means_lock)
-        obj.enter_context(obj.table_counts_lock)
-        return obj
-
-
-class SharedArray:
-    """
-    Small wrapper for a multiprocessing shared-memory array that will be used as the
-    in-memory storage for a numpy array. This allows a numpy array to be easily shared
-    between processes.
-
-    The array has a lock that should be acquired before reading or writing the numpy
-    array. The lock is not enforced automatically: you should always enclose any
-    numpy operations that access the shared-memory array in:
-
-       with arr.lock:
-           # Some numpy operations on arr.np: the numpy array backed by the shared memory
-           arr.np[0] = 1
-
-    """
-    def __init__(self, array, shape, lock, dtype):
-        self.array = array
-        self.shape = shape
-        self.np = np.frombuffer(self.array, dtype=dtype).reshape(self.shape)
-        # We use our own lock, not the mp.Array one, so that we can ensure all parameters are updated in a single op
-        self.lock = lock
-
-    @staticmethod
-    def create(shape, dtype="float"):
-        if dtype == "float":
-            ctype = "d"
-            np_type = np.float64
-        elif dtype == "int":
-            ctype = "l"
-            np_type = np.int64
-        else:
-            raise ValueError("unknown type '{}'".format(dtype))
-
-        lock = mp.Lock()
-        # Allocate shared memory to sit behind the numpy array
-        if type(shape) is int:
-            size = shape
-        else:
-            size = functools.reduce(mul, shape)
-        array = sharedctypes.RawArray(ctype, size)
-        return SharedArray(array, shape, lock, np_type)
-
-    def __getstate__(self):
+    def log_multivariate_tdensity(self, x, table_id):
         """
-        Arrays get pickled to be sent between processes. We ensure that the same
-        shared array is used on the other side.
+        Gaussian likelihood for a table-embedding pair when using Cholesky decomposition.
 
         """
-        return self.array, self.shape, self.lock, self.np.dtype
+        # Make sure the main process doesn't update the parameters in the middle of the computation
+        with self.param_lock:
+            # Copy these values so we can release the lock and use a consistent state
+            count = self.table_counts[table_id].copy()
+            table_cholesky_ltriangular_mat = self.table_cholesky_ltriangular_mat[table_id].copy()
+            log_determinant = self.log_determinants[table_id].copy()
+            table_mean = self.table_means[table_id].copy()
 
-    def __setstate__(self, state):
-        self.__init__(*state)
+        k_n = self.kappa + count
+        nu_n = self.nu + count
+        scaleTdistrn = np.sqrt((k_n + 1.) / (k_n * (nu_n - self.embedding_size + 1.)))
+        nu = self.nu + count - self.embedding_size + 1.
+        # Since I am storing lower triangular matrices, it is easy to calculate (x-\mu)^T\Sigma^-1(x-\mu)
+        # therefore I am gonna use triangular solver
+        # first calculate (x-mu)
+        x_minus_mu = x - table_mean
+        # Now scale the lower tringular matrix
+        ltriangular_chol = scaleTdistrn * table_cholesky_ltriangular_mat
+        solved = solve_triangular(ltriangular_chol, x_minus_mu)
+        # Now take xTx (dot product)
+        val = (solved ** 2.).sum()
+
+        logprob = gammaln((nu + self.embedding_size) / 2.) - \
+                  (
+                          gammaln(nu / 2.) +
+                          self.embedding_size / 2. * (np.log(nu) + np.log(math.pi)) +
+                          log_determinant +
+                          (nu + self.embedding_size) / 2. * np.log(1. + val / nu)
+                  )
+        return logprob
+
+    def log_multivariate_tdensity_tables(self, x):
+        """
+        Gaussian likelihood for a table-embedding pair when using Cholesky decomposition.
+        This version computes the likelihood for all tables in parallel.
+
+        """
+        lprobs = np.zeros(self.num_tables, dtype=np.float32)
+        for i in range(self.num_tables):
+            lprobs[i] = self.log_multivariate_tdensity(x, i)
+        return lprobs
+
+
+class SamplingDiagnostics:
+    """
+    Keeps track of a few statistics about the sampling process that can be useful to
+    display between iterations to give an idea of what's happening during sampling.
+    This is mainly for debugging, but is quite a useful thing to output to give
+    an idea of what the sampler's doing in each iteration.
+
+    """
+    def __init__(self):
+        self._select_pr_sum = 0.
+        self._select_pr_uses = 0
+        self._pr_accepted = 0
+        self._acceptance_sum = 0.
+        self._acceptance_uses = 0
+        self._accepted_total = 0
+
+    def log_select_pr(self, selected, select_pr):
+        self._select_pr_sum += float(select_pr)
+        self._select_pr_uses += 1
+        if selected:
+            self._pr_accepted += 1
+
+    def log_acceptance(self, accepted, acceptance):
+        self._acceptance_sum += float(acceptance)
+        self._acceptance_uses += 1
+        if accepted:
+            self._accepted_total += 1
+
+    def acceptance_used(self):
+        return self._acceptance_uses > 0
+
+    def mean_acceptance(self):
+        return self._acceptance_sum / self._acceptance_uses
+
+    def acceptance_rate(self):
+        return float(self._accepted_total) / self._acceptance_uses
+
+    def mean_select_pr(self):
+        return self._select_pr_sum / self._select_pr_uses
+
+    def select_pr_rate(self):
+        return float(self._pr_accepted) / self._select_pr_uses
